from typing import Optional, cast

from datahub.configuration.common import ConfigModel
from datahub.emitter.mce_builder import Aspect
from datahub.ingestion.api.common import PipelineContext
from datahub.ingestion.transformer.dataset_transformer import (
    DatasetOwnershipTransformer,
)
from datahub.metadata.schema_classes import OwnershipClass


class ClearDatasetOwnershipConfig(ConfigModel):
    pass


class SimpleRemoveDatasetOwnership(DatasetOwnershipTransformer):
    """Transformer that clears all owners on each dataset."""

    def __init__(self, config: ClearDatasetOwnershipConfig, ctx: PipelineContext):
        super().__init__()

    @classmethod
    def create(
        cls, config_dict: dict, ctx: PipelineContext
    ) -> "SimpleRemoveDatasetOwnership":
        config = ClearDatasetOwnershipConfig.parse_obj(config_dict)
        return cls(config, ctx)

    def transform_aspect(
        self, entity_urn: str, aspect_name: str, aspect: Optional[Aspect]
    ) -> Optional[Aspect]:
        in_ownership_aspect = cast(OwnershipClass, aspect)
        if in_ownership_aspect is None:
            return cast(Aspect, in_ownership_aspect)

        in_ownership_aspect.owners = []

        return cast(Aspect, in_ownership_aspect)



from typing import Dict, Optional, Union

from azure.identity import ClientSecretCredential
from azure.storage.filedatalake import DataLakeServiceClient, FileSystemClient
from pydantic import Field, root_validator

from datahub.configuration import ConfigModel
from datahub.configuration.common import ConfigurationError


class AdlsSourceConfig(ConfigModel):
    """
    Common Azure credentials config.

    https://docs.microsoft.com/en-us/azure/storage/blobs/data-lake-storage-directory-file-acl-python
    """

    base_path: str = Field(
        default="/",
        description="Base folder in hierarchical namespaces to start from.",
    )
    container_name: str = Field(
        description="Azure storage account container name.",
    )
    account_name: str = Field(
        description="Name of the Azure storage account.  See [Microsoft official documentation on how to create a storage account.](https://docs.microsoft.com/en-us/azure/storage/blobs/create-data-lake-storage-account)",
    )
    account_key: Optional[str] = Field(
        description="Azure storage account access key that can be used as a credential. **An account key, a SAS token or a client secret is required for authentication.**",
        default=None,
    )
    sas_token: Optional[str] = Field(
        description="Azure storage account Shared Access Signature (SAS) token that can be used as a credential. **An account key, a SAS token or a client secret is required for authentication.**",
        default=None,
    )
    client_secret: Optional[str] = Field(
        description="Azure client secret that can be used as a credential. **An account key, a SAS token or a client secret is required for authentication.**",
        default=None,
    )
    client_id: Optional[str] = Field(
        description="Azure client (Application) ID required when a `client_secret` is used as a credential.",
        default=None,
    )
    tenant_id: Optional[str] = Field(
        description="Azure tenant (Directory) ID required when a `client_secret` is used as a credential.",
        default=None,
    )

    def get_abfss_url(self, folder_path: str = "") -> str:
        if not folder_path.startswith("/"):
            folder_path = f"/{folder_path}"
        return f"abfss://{self.container_name}@{self.account_name}.dfs.core.windows.net{folder_path}"

    def get_filesystem_client(self) -> FileSystemClient:
        return self.get_service_client().get_file_system_client(self.container_name)

    def get_service_client(self) -> DataLakeServiceClient:
        return DataLakeServiceClient(
            account_url=f"https://{self.account_name}.dfs.core.windows.net",
            credential=self.get_credentials(),
        )

    def get_credentials(
        self,
    ) -> Union[Optional[str], ClientSecretCredential]:
        if self.client_id and self.client_secret and self.tenant_id:
            return ClientSecretCredential(
                tenant_id=self.tenant_id,
                client_id=self.client_id,
                client_secret=self.client_secret,
            )
        return self.sas_token if self.sas_token is not None else self.account_key

    @root_validator()
    def _check_credential_values(cls, values: Dict) -> Dict:
        if (
            values.get("account_key")
            or values.get("sas_token")
            or (
                values.get("client_id")
                and values.get("client_secret")
                and values.get("tenant_id")
            )
        ):
            return values
        raise ConfigurationError(
            "credentials missing, requires one combination of account_key or sas_token or (client_id and client_secret and tenant_id)"
        )



from typing import Dict, Optional, Set

from pydantic import validator
from pydantic.fields import Field

from datahub.configuration.common import ConfigModel, ConfigurationError
from datahub.metadata.schema_classes import FabricTypeClass

DEFAULT_ENV = FabricTypeClass.PROD

# Get all the constants from the FabricTypeClass. It's not an enum, so this is a bit hacky but works.
ALL_ENV_TYPES: Set[str] = set(
    [value for name, value in vars(FabricTypeClass).items() if not name.startswith("_")]
)


class PlatformSourceConfigBase(ConfigModel):
    """
    Any source that connects to a platform should inherit this class
    """

    platform: Optional[str] = Field(
        default=None, description="The platform that this source connects to"
    )

    platform_instance: Optional[str] = Field(
        default=None,
        description="The instance of the platform that all assets produced by this recipe belong to",
    )


class EnvBasedSourceConfigBase(ConfigModel):
    """
    Any source that produces dataset urns in a single environment should inherit this class
    """

    env: str = Field(
        default=DEFAULT_ENV,
        description="The environment that all assets produced by this connector belong to",
    )

    @validator("env")
    def env_must_be_one_of(cls, v: str) -> str:
        if v.upper() not in ALL_ENV_TYPES:
            raise ConfigurationError(f"env must be one of {ALL_ENV_TYPES}, found {v}")
        return v.upper()


class DatasetSourceConfigBase(PlatformSourceConfigBase, EnvBasedSourceConfigBase):
    """
    Any source that is a primary producer of Dataset metadata should inherit this class
    """


class DatasetLineageProviderConfigBase(EnvBasedSourceConfigBase):
    """
    Any non-Dataset source that produces lineage to Datasets should inherit this class.
    e.g. Orchestrators, Pipelines, BI Tools etc.
    """

    platform_instance_map: Optional[Dict[str, str]] = Field(
        default=None,
        description="A holder for platform -> platform_instance mappings to generate correct dataset urns",
    )



import re
from typing import Dict, List, Optional, Union
from urllib.parse import urlparse

from pydantic import Field, root_validator, validator

from datahub.configuration.common import AllowDenyPattern, ConfigurationError
from datahub.configuration.source_common import DEFAULT_ENV, DatasetSourceConfigBase
from datahub.ingestion.source.state.stale_entity_removal_handler import (
    StatefulStaleMetadataRemovalConfig,
)
from datahub.ingestion.source.state.stateful_ingestion_base import (
    StatefulIngestionConfigBase,
)
from datahub.utilities import config_clean


class PulsarSourceStatefulIngestionConfig(StatefulStaleMetadataRemovalConfig):
    """
    Specialization of the StatefulStaleMetadataRemovalConfig to add custom config.
    This will be used to override the stateful_ingestion config param of StatefulIngestionConfigBase
    in the PulsarSourceConfig.
    """

    _entity_types: List[str] = Field(default=["topic"])


def _is_valid_hostname(hostname: str) -> bool:
    """
    Loosely ascii hostname validation. A hostname is considered valid when the total length does not exceed 253
    characters, contains valid characters and are max 63 octets per label.
    """
    if len(hostname) > 253:
        return False
    # Hostnames ending on a dot are valid, if present strip exactly one
    if hostname[-1] == ".":
        hostname = hostname[:-1]
    allowed = re.compile(r"(?!-)[A-Z\d-]{1,63}(?<!-)$", re.IGNORECASE)
    return all(allowed.match(x) for x in hostname.split("."))


class PulsarSourceConfig(StatefulIngestionConfigBase, DatasetSourceConfigBase):
    env: str = DEFAULT_ENV

    web_service_url: str = Field(
        default="http://localhost:8080", description="The web URL for the cluster."
    )
    timeout: int = Field(
        default=5,
        description="Timout setting, how long to wait for the Pulsar rest api to send data before giving up",
    )
    # Mandatory for oauth authentication
    issuer_url: Optional[str] = Field(
        default=None,
        description="The complete URL for a Custom Authorization Server. Mandatory for OAuth based authentication.",
    )
    client_id: Optional[str] = Field(
        default=None, description="The application's client ID"
    )
    client_secret: Optional[str] = Field(
        default=None, description="The application's client secret"
    )
    # Mandatory for token authentication
    token: Optional[str] = Field(
        default=None,
        description="The access token for the application. Mandatory for token based authentication.",
    )
    verify_ssl: Union[bool, str] = Field(
        default=True,
        description="Either a boolean, in which case it controls whether we verify the server's TLS certificate, or a string, in which case it must be a path to a CA bundle to use.",
    )
    # By default, allow all topics and deny the pulsar system topics
    tenant_patterns: AllowDenyPattern = Field(
        default=AllowDenyPattern(allow=[".*"], deny=["pulsar"]),
        description="List of regex patterns for tenants to include/exclude from ingestion. By default all tenants are allowed.",
    )
    namespace_patterns: AllowDenyPattern = Field(
        default=AllowDenyPattern(allow=[".*"], deny=["public/functions"]),
        description="List of regex patterns for namespaces to include/exclude from ingestion. By default the functions namespace is denied.",
    )
    topic_patterns: AllowDenyPattern = Field(
        default=AllowDenyPattern(allow=[".*"], deny=["/__.*$"]),
        description="List of regex patterns for topics to include/exclude from ingestion. By default the Pulsar system topics are denied.",
    )
    exclude_individual_partitions: bool = Field(
        default=True,
        description="Extract each individual partitioned topic. e.g. when turned off a topic with 100 partitions will result in 100 Datasets.",
    )

    tenants: List[str] = Field(
        default=[],
        description="Listing all tenants requires superUser role, alternative you can set a list of tenants you want to scrape using the tenant admin role",
    )

    domain: Dict[str, AllowDenyPattern] = Field(
        default_factory=dict, description="Domain patterns"
    )

    stateful_ingestion: Optional[PulsarSourceStatefulIngestionConfig] = Field(
        default=None, description="see Stateful Ingestion"
    )

    oid_config: dict = Field(
        default_factory=dict, description="Placeholder for OpenId discovery document"
    )

    @validator("token")
    def ensure_only_issuer_or_token(
        cls, token: Optional[str], values: Dict[str, Optional[str]]
    ) -> Optional[str]:
        if token is not None and values.get("issuer_url") is not None:
            raise ConfigurationError(
                "Expected only one authentication method, either issuer_url or token."
            )
        return token

    @validator("client_secret", always=True)
    def ensure_client_id_and_secret_for_issuer_url(
        cls, client_secret: Optional[str], values: Dict[str, Optional[str]]
    ) -> Optional[str]:
        if values.get("issuer_url") is not None and (
            client_secret is None or values.get("client_id") is None
        ):
            raise ConfigurationError(
                "Missing configuration: client_id and client_secret are mandatory when issuer_url is set."
            )
        return client_secret

    @validator("web_service_url")
    def web_service_url_scheme_host_port(cls, val: str) -> str:
        # Tokenize the web url
        url = urlparse(val)

        if url.scheme not in ["http", "https"]:
            raise ConfigurationError(
                f"Scheme should be http or https, found {url.scheme}"
            )

        if not _is_valid_hostname(url.hostname.__str__()):
            raise ConfigurationError(
                f"Not a valid hostname, hostname contains invalid characters, found {url.hostname}"
            )

        return config_clean.remove_trailing_slashes(val)

    @root_validator
    def validate_platform_instance(cls: "PulsarSourceConfig", values: Dict) -> Dict:
        stateful_ingestion = values.get("stateful_ingestion")
        if (
            stateful_ingestion
            and stateful_ingestion.enabled
            and not values.get("platform_instance")
        ):
            raise ConfigurationError(
                "Enabling Pulsar stateful ingestion requires to specify a platform instance."
            )
        return values
